# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2019 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Members API."""

from __future__ import absolute_import, print_function

from collections import defaultdict

from flask import current_app, url_for
from invenio_accounts.models import User
from invenio_db import db
# from invenio_communities.utils import send_invitation_email
from invenio_records.api import Record as RecordBaseAPI
from werkzeug.local import LocalProxy

from invenio_communities.members.models import \
    CommunityMember as CommunityMemberModel
from invenio_communities.members.models import CommunityMemberRole, \
    CommunityMemberStatus
from invenio_communities.requests.api import RequestBase

# from invenio_communities.api import Community


Community = LocalProxy(
    lambda: current_app.extensions['invenio-communities'].community_cls)


class CommunityMemberRequest(RequestBase):
    """Request class for Member objects."""

    TYPE = 'community-member-request'

    community_member_cls = LocalProxy(lambda: CommunityMember)

    schema = {
        "type": {"type": "string"},
        "state": {"type": "string"},
        "created_by": {"type": "int"},
    }

    @classmethod
    def create(cls, owner, id_=None, **kwargs):
        """Create a CommunityMemberRequest."""
        data = {
            'type': cls.TYPE,
            'state': 'OPEN',
            'created_by': owner.id,
            **kwargs,
        }

        model = cls.model_cls(
            owner_id=owner.id,
            json=data,
            id=id_,
        )
        db.session.add(model)
        return cls(data, model=model)

    @property
    def owner(self):
        """Get owner."""
        if not getattr(self, '_owner', None) and self.model.owner_id:
            self._owner = User.query.get(self.model.owner_id)
        return getattr(self, '_owner', None)

    @property
    def community_member(self):
        """Get request's community record relatinship."""
        if not getattr(self, '_community_member', None):
            self._community_member = \
                self.community_member_cls.get_by_request_id(request_id=self.id)
        return self._community_member

    @property
    def community(self):
        """Get request community."""
        return self.community_member.community

    @property
    def user(self):
        """Get request record."""
        return self.community_member.user

    def as_dict(self):
        """Return the metadata of the request as a dictionary."""
        if self.is_invite:
            request_type = 'invitation'
        else:
            request_type = 'request'
        return {
            'id': self.id,
            'created': self.created,
            'updated': self.updated,
            'comments': [
                {
                    'message': c.message,
                    'created_by': c.created_by,
                    'created': c.created,
                } for c in self.comments
            ],
            'request_type': request_type,
            'email': self.community_member.invitation_id
        }

    @property
    def is_invite(self):
        """Returns true or false depending on the request direction."""
        return bool(self.community_member.invitation_id)

    @property
    def is_closed(self):
        """Returns true or false depending on the state of the request."""
        return self['state'] == 'CLOSED'

    def close_request(self):
        """Close the request after it has been succesfully handled."""
        self['state'] = 'CLOSED'


class ModelProxyProperty(object):
    """Class for initializing property like objects."""

    def __init__(self, name):
        """."""
        self.name = name

    def __get__(self, obj, objtype=None):
        """."""
        return getattr(obj.model, self.name)

    def __set__(self, obj, val):
        """."""
        setattr(obj.model, self.name, val)

class CommunityMember(RecordBaseAPI):
    """Community Member class."""

    model_cls = CommunityMemberModel
    community_cls = Community

    @property
    def request(self):
        """Community member request."""
        if self.model and self.model.request:
            return CommunityMemberRequest(
                self.model.request.json, self.model.request)
        else:
            return None

    @property
    def community(self):
        """Get community."""
        if not getattr(self, '_community', None):
            self._community = self.community_cls.get_record(
                self.model.community_pid.object_uuid)
        return self._community

    @property
    def user(self):
        """Get user."""
        if not getattr(self, '_user', None) and self.model.user_id:
            self._user = User.query.get(self.model.user_id)
        return getattr(self, '_user', None)

    role = ModelProxyProperty('role')

    status = ModelProxyProperty('status')

    invitation_id = ModelProxyProperty('invitation_id')

    user_id = ModelProxyProperty('user_id')

    @classmethod
    def create(cls, community, role, user=None, request=None, invitation_id=None, status=None, data=None):
        """Create a community member relationship."""
        assert invitation_id or user
        data = data or {}
        request_id = request.id if request else None
        model = cls.model_cls.create(
            community_pid_id=community.pid.id,
            user_id=user.id if user else None,
            invitation_id=invitation_id,
            request_id=request_id,
            role=role,
            status=status,
            json=data,
        )
        obj = cls(data, model=model)
        obj._community = community
        if user:
            obj._user = user
        return obj

    def delete(self):
        """Delete community member relationship."""
        return self.model.delete(self.model)

    @classmethod
    def get_by_ids(cls, community_pid, user_id):
        """Get by community and user ID."""
        model = CommunityMemberModel.query.filter_by(
            community_pid_id=community_pid.id,
            user_id=user_id,
        ).one_or_none()
        if not model:
            return None
        return cls(model.json, model=model)

    @classmethod
    def get_by_id(cls, membership_id):
        """Get by community and user ID."""
        model = CommunityMemberModel.query.get(membership_id)
        if not model:
            return None
        return cls(model.json, model=model)

    @classmethod
    def get_by_request_id(cls, request_id):
        """Get by request ID."""
        model = CommunityMemberModel.query.filter_by(
            request_id=request_id
        ).one_or_none()
        if not model:
            return None
        return cls(model.json, model=model)

    # TODO: Remove
    def dump_links(self):
        actions = ['comment', 'accept', 'reject']
        links = {
            "self": url_for(
                'invenio_communities_members.community_requests_api',
                pid_value=self.community.pid.pid_value,
                membership_id=str(self.id)
            )
        }
        for action in actions:
            links[action] = url_for(
                'invenio_communities_members.community_requests_handling_api',
                pid_value=self.community.pid.pid_value,
                membership_id=str(self.id),
                action=action
            )
        return links

    def as_dict(self, include_requests=False):
        res = {
            'id': str(self.id),
            'status': str(self.status.name.lower()),
            'role': str(self.role.name.lower()),
            'user_id': self.user.id if self.user else None,
            'username': self.user.profile._displayname if self.user and self.user.profile else None,
            # TODO: Shouldn't be visible publicly. This data should be
            # removed/cleaned in the controller
            'email': self.invitation_id,
            # TODO: Generate these in the view
            #'links': self.dump_links(),
        }
        if self.request:
            if include_requests:
                res['request'] = self.request.as_dict()
            else:
                res['request_id'] = self.request.id
        return res


class CommunityMembersCollection:
    """Iterator for Community Members."""

    community_member_cls = CommunityMember

    def __init__(self, community, _query=None):
        """Initialize the iterator by providing the community."""
        self.community = community
        self._query = _query or CommunityMemberModel.query.filter_by(
            community_pid_id=self.community.pid.id
        ).order_by(
            CommunityMemberModel.created.desc()
        )

    def __len__(self):
        """Get number of community members."""
        return self._query.count()

    def __iter__(self):
        """Iterate over the DB query."""
        self._it = iter(self._query)
        return self

    def filter(self, **conditions):
        """Filter the members with additional conditions."""
        new_query = self._query.filter_by(**conditions)
        return self.__class__(self.community, _query=new_query)

    def __next__(self):
        """Get next community member item."""
        obj = next(self._it)
        return self.community_member_cls(obj.json, model=obj)

    def __contains__(self, user):
        """Method that checks if a user is part of the community members."""
        if isinstance(user, User):
            user_id = user.id
        elif isinstance(user, int):
            user_id = user
        return bool(self._query.filter_by(user_id=user_id).count())

    def __getitem__(self, user_id):
        """Get a specific community member by user ID."""
        return self.community_member_cls.get_by_ids(
            self.community.pid, user_id)

    def add(self, request, status=CommunityMemberStatus.PENDING, user=None,
            role=CommunityMemberRole.MEMBER, invitation_id=None):
        """Add member to the community."""
        return self.community_member_cls.create(
            community=self.community, request=request, role=role, user=user,
            status=status, invitation_id=invitation_id)

    def remove(self, user):
        """Remove a member from the community."""
        community_member = self[user.id]
        return community_member.delete()

    def paginate(self, page=1, size=20):
        pagination = self._query.paginate(page=int(page), per_page=int(size))
        return [self.community_member_cls(i.json, model=i)
                for i in pagination.items]

    def as_dict(self, include_requests=False, result_iterator=None):
        res = []
        result_iterator = result_iterator or self
        for community_member in result_iterator:
            res.append(community_member.as_dict(
                include_requests=include_requests)
            )
        #TODO add aggregations
        res = {"hits": {
            "total": self._query.count(),
            "hits": res
        }}
        return res

    def aggregate(self, key):
        return self._query.group_by(key).count()

    #TODO check this
    def get_user_membership(self, user_id, **kwargs):
        community_members = self.filter(
            user_id=user_id,
            **kwargs)
        membership_obj = community_members._query.one_or_none()
        if not membership_obj:
            return None
        return self.community_member_cls(
            membership_obj.json, model=membership_obj)


class CommunityMembersMixin:

    community_members_iter_cls = CommunityMembersCollection

    @property
    def members(self):
        return self.community_members_iter_cls(self)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community records errors."""

from ..errors import CommunitiesError


class CommunityMemberAlreadyExists(CommunitiesError):
    """Community membership already exists error."""

    def __init__(self, user_id, pid_id, invitation_id):
        """Initialize Exception."""
        self.user_id = user_id
        self.pid_id = pid_id
        self.invitation_id = invitation_id



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019 CERN.
# Copyright (C) 2019 Northwestern University.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Permissions for Invenio RDM Records."""
from collections import namedtuple
from functools import partial

from flask_principal import UserNeed
from invenio_access.permissions import any_user
from invenio_records_permissions.generators import Generator
from invenio_records_permissions.policies import BasePermissionPolicy

_CommunityMemberNeed = namedtuple(
    'CommunityMemberNeed', ['method', 'community', 'role'])
CommunityMemberNeed = partial(_CommunityMemberNeed, "community_member")


class CommunityMember(Generator):
    """Generator for community member need."""

    def __init__(self, roles):
        """."""
        self.roles = roles

    def needs(self, comid=None, **kwargs):
        """Returns community member need."""
        return [CommunityMemberNeed(comid.id, r) for r in self.roles]


class CommunityMembershipOwner(Generator):

    def needs(self, community_member=None, **kwargs):
        """Returns community member need."""
        if community_member.user_id:
            return [UserNeed(community_member.user_id)]
        return []


class AnyUserIfUnresolvedInvite(Generator):

    def needs(self, community_member=None, **kwargs):
        """Skip check at this stage if the membership is an invitation."""
        if community_member.request.is_invite:
            return [any_user]
        return []


class CommunityMemberIfRequest(Generator):

    def __init__(self, roles):
        self.roles = roles

    def needs(self, comid=None, community_member=None, **kwargs):
        """If its a request only defined community roles can handle it."""
        if not community_member.request.is_invite:
            return [CommunityMemberNeed(comid.id, r) for r in self.roles]
        return []


class AnyUserIfOpenMemberPolicy(Generator):

    def needs(self, community=None, **kwargs):
        """If the member policy is open then anyone can request to join."""
        if community['member_policy'] == 'open':
            return [any_user]
        return []


class CommunityMemberPermissionPolicy(BasePermissionPolicy):
    """Permission policy for member functionality."""
    can_create_membership = [
        CommunityMember(['admin']),
        AnyUserIfOpenMemberPolicy()
    ]
    can_list_accepted_members = [CommunityMember(['admin', 'curator', 'member'])]
    can_list_pending_members = [CommunityMember(['admin'])]
    can_list_rejected_members = [CommunityMember(['admin'])]
    can_list_all_members = [CommunityMember(['admin'])]

    can_modify_membership = [CommunityMember(['admin'])]

    can_delete_membership = [
        CommunityMember(['admin']),
        CommunityMembershipOwner(),
    ]

    can_handle_request = [
        AnyUserIfUnresolvedInvite(),
        CommunityMemberIfRequest(['admin']),
    ]

    can_get_membership = [
        CommunityMemberIfRequest(['admin']),
        CommunityMembershipOwner(),
        AnyUserIfUnresolvedInvite()
    ]



def is_permitted_action(action, **kwargs):
    """Helper function to apply a permission check."""
    return CommunityMemberPermissionPolicy(
        action=action, **kwargs).can()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community module tests."""

import uuid

import pytest
from flask import url_for
from invenio_accounts.testutils import create_test_user, login_user_via_session
from invenio_indexer.api import RecordIndexer
from invenio_pidstore.models import PersistentIdentifier, PIDStatus
from invenio_records.api import Record

from invenio_communities.api import Community
from invenio_communities.indexer import CommunityIndexer


@pytest.fixture
def record(db, es, users):
    _, record_owner, _ = users
    record = Record.create({
        'title': 'Title',
        '_owners': [record_owner.id],
    })
    recid = PersistentIdentifier.create(
        pid_type='recid', pid_value='12345',
        object_uuid=record.id, object_type='rec',
        status='R')
    db.session.commit()
    RecordIndexer().index_by_id(str(record.id))
    yield recid, record


@pytest.fixture
def community(db, es, users):
    community_owner, _, _ = users
    community_id = str(uuid.uuid4())
    comid = PersistentIdentifier.create(
        pid_type='comid', pid_value='biosyslit',
        object_uuid=community_id, object_type='com',
        status='R')
    community = Community.create({
        'id': 'comm_id',
        'title': 'Title',
        'type': 'topic',
        'created_by': community_owner.id,
    }, id_=community_id)
    db.session.commit()
    CommunityIndexer().index_by_id(str(community.id))
    yield comid, community


def test_simple_flow(db, es_clear, community, record, client, users):
    """Test basics operations on records."""
    # TODO
    # /api/communities/<comid>/requests/records/
    comid, community = community
    recid, record = record
    _, record_owner, _ = users
    login_user_via_session(client, user=record_owner)
    community_records_list_url = url_for(
        'invenio_communities_records.requests_list', pid_value=comid.pid_value)
    resp = client.post(community_records_list_url, json={
        'pid': recid.pid_value,
        'record_pid': recid.pid_value,
    })

    # create

    # read

    # update

    # delete

    pass


def test_records_permissions(db, es_clear, community, record, client, users):
    """Test permissions on records."""
    # TODO
    pass


def test_records_validation(db, es_clear, community, record, client, users):
    """Test records validation."""
    # TODO
    pass


def test_records_search(db, es_clear, community, record, client, users):
    """Test records search."""
    # TODO
    pass



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprint definitions."""

from __future__ import absolute_import, print_function

from functools import partial, wraps

from flask import Blueprint, abort, jsonify, render_template, request
from flask.views import MethodView
from flask_login import current_user, login_required
from invenio_accounts.models import User
from invenio_db import db
from webargs import fields, validate

from invenio_communities.records.api import CommunityRecordsCollection

from ..token import MembershipTokenSerializer
from ..utils import comid_url_converter, send_invitation_email
from ..views import pass_community, use_kwargs
from .api import CommunityMember, CommunityMemberRequest
from .errors import CommunityMemberAlreadyExists
from .models import CommunityMemberRole, CommunityMemberStatus
from .permissions import CommunityMemberPermissionPolicy, is_permitted_action

api_blueprint = Blueprint(
        'invenio_communities_members',
        __name__,
        template_folder="../templates",
    )


def pass_membership(func=None, view_arg_name='membership_id'):
    """Decorator to retrieve community membership."""
    if func is None:
        return partial(pass_membership, view_arg_name=view_arg_name)

    @wraps(func)
    def inner(*args, **kwargs):
        del kwargs[view_arg_name]
        membership_id = request.view_args[view_arg_name]
        community_member = CommunityMember.get_by_id(membership_id)
        if community_member is None:
            abort(404)
        return func(*args, community_member=community_member, **kwargs)
    return inner


def community_permission(
        action=None,
        error_message='You are missing the permissions to'
                      'access this information.',
        error_code=403):
    """Wrapper to apply a permission check on a view."""
    def wrapper_function(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not CommunityMemberPermissionPolicy(action=action, **kwargs):
                abort(error_code, error_message)
            return func(*args, **kwargs)
        return wrapper
    return wrapper_function


member_role_validator = validate.OneOf(
    [c.name.lower() for c in CommunityMemberRole])

member_status_validator = validate.OneOf(
    [c.name.lower() for c in CommunityMemberStatus])


class ListResource(MethodView):
    """Resource for creating, listing and removing community memberships."""

    post_args = {
        'role': fields.Str(
            location='json',
            required=False,
            validate=member_role_validator,
        ),
        'email': fields.Email(
            location='json',
            required=False,
        ),
        'comment': fields.Str(
            location='json',
            required=False,
        ),
    }

    @use_kwargs(post_args)
    @pass_community
    @login_required
    @community_permission('create_membership')
    def post(self, comid=None, community=None, email=None, role=None,
             comment=None, **kwargs):
        """Join a community or invite a user to it."""
        request_user = User.query.get(int(current_user.get_id()))
        # TODO: see if there's a marshmallow field for this
        if role:
            role = CommunityMemberRole.from_str(role)
        mem_req = CommunityMemberRequest.create(owner=request_user)
        if request_user in community.members.filter(
                role=CommunityMemberRole.ADMIN,
                status=CommunityMemberStatus.ACCEPTED):
            # invitation
            if not email:
                abort(400, 'Email is a required field.')
            try:
                comm_mem = community.members.add(
                    mem_req,
                    status=CommunityMemberStatus.PENDING,
                    role=role,
                    invitation_id=email,
                )
            except CommunityMemberAlreadyExists:
                abort(400, 'This is an already existing relationship.')
        else:
            # request to join
            if email:
                abort(403, 'You need to be a community administrator to '
                           'invite members to this community')
            if role:
                abort(403, 'Selecting a role is not possible when requesting'
                           ' to join a community')
            try:
                comm_mem = community.members.add(
                    mem_req,
                    status=CommunityMemberStatus.PENDING,
                    user=request_user
                )
            except CommunityMemberAlreadyExists:
                abort(400, 'This is an already existing relationship.')

        if comment:
            mem_req.add_comment(request_user.id, comment)

        if email:
            mts = MembershipTokenSerializer()
            inv_id = comm_mem.invitation_id
            token = mts.create_token(inv_id)
            send_invitation_email(comm_mem, [email], community, token)
        else:
            emails = []
            admins = community.members.filter(
                role=CommunityMemberRole.ADMIN,
                status=CommunityMemberStatus.ACCEPTED)
            for admin in admins:
                emails.append(admin.user.email)
                send_invitation_email(comm_mem, emails, community)

        db.session.commit()
        return comm_mem.as_dict(), 201


class ListMembersResource(MethodView):
    """Resource for creating, listing and removing community memberships."""

    get_args = {
        #TODO make list when we manage to integrate the in operator
        'status': fields.Str(
            location='querystring',
            required=False,
            validate=member_status_validator
        ),
        'role': fields.List(fields.Raw(
            location='querystring',
            required=False
        )
        ),
        'size': fields.Raw(
            location='querystring',
            required=False
        ),
        'page': fields.Raw(
            location='querystring',
            required=False
        )
    }

    @use_kwargs(get_args)
    @pass_community
    @login_required
    def get(self, comid=None, community=None,
            status=None, role=None, page=1, size=20):
        """List the community members."""
        if not is_permitted_action(
            'list_{}_members'.format(status),
                comid=comid):
            abort(403, 'You are missing the permissions to'
                       'access this information.')
        status = CommunityMemberStatus.from_str(status)

        members = community.members
        members = members.filter(status=status)
        if status != 'accepted':
            include_requests = True
        else:
            include_requests = False
        if role:
            members = members.filter(role=role)
        member_page = members.paginate(page=page, size=size)
        return members.as_dict(
            include_requests=include_requests,
            result_iterator=member_page
            ), 200


class MembershipRequestResource(MethodView):
    """Resource to view and handle membership requests."""

    get_args = {
        'token': fields.Raw(
            location='query',
            required=False,
        )}

    @use_kwargs(get_args)
    @pass_community
    @pass_membership
    @login_required
    @community_permission('get_membership')
    def get(
            self, comid=None, community=None, token=None,
            community_member=None):
        """Get a membership along with it's request information."""
        mts = MembershipTokenSerializer()
        inv_id = community_member.invitation_id

        if community_member.request.is_invite \
            and (not token or not mts.validate_token(
                    token, expected_value=inv_id)):
            abort(404)
        # TODO links are currently missing.
        return jsonify(community_member.as_dict(include_requests=True)), 200

    put_args = {
        'role': fields.Raw(
            location='json',
            required=True,
            validate=member_role_validator,
        )
    }

    @use_kwargs(put_args)
    @pass_community
    @pass_membership
    @login_required
    @community_permission('modify_membership')
    def put(self, comid=None, community=None,
            role=None, community_member=None):
        """Modify a membership role."""
        community_admins = community.members.filter(
            role=CommunityMemberRole.ADMIN,
            status=CommunityMemberStatus.ACCEPTED)

        if len(community_admins) == 1 and \
                community_member.role == CommunityMemberRole.ADMIN:
            abort(400, 'Cannot remove last community Administrator.')

        community_member.role = CommunityMemberRole.from_str(role)
        db.session.commit()
        # return jsonify(community_member.as_dict()), 200
        return 'Succesfully modified invitation.', 200

    del_args = {
        'token': fields.Raw(
            location='json',
            required=False,
        )
    }

    @use_kwargs(del_args)
    @pass_community
    @pass_membership
    @login_required
    @community_permission('delete_membership')
    def delete(self, comid=None, community=None, community_member=None):
        """Cancel (remove) a membership request."""
        # TODO dont allow last admin to leave
        community_admins = community.members.filter(
            role=CommunityMemberRole.ADMIN,
            status=CommunityMemberStatus.ACCEPTED)
        if len(community_admins) == 1 and \
                community_member.role == CommunityMemberRole.ADMIN:
            abort(400, 'Cannot remove last community Administrator.')

        community_member.delete()
        db.session.commit()
        return 'Succesfully removed membership.', 204


class MembershipRequestHandlingResource(MethodView):
    """Resource to view and handle membership requests."""

    post_args = {
        'role': fields.Raw(
            location='json',
            required=False,
            validate=member_role_validator,
        ),
        'token': fields.Raw(
            location='json',
            required=False,
        ),
        'comment': fields.Raw(
            location='json',
            required=False,
        )
    }

    @use_kwargs(post_args)
    @pass_community
    @pass_membership
    @login_required
    @community_permission('handle_request')
    def post(self, comid=None, community=None, community_member=None,
             action=None, role=None, comment=None, token=None):
        """Add a comment."""
        request_user = User.query.get(int(current_user.get_id()))
        #
        # Controller
        #
        if community_member.request.is_closed and \
                action != 'comment':
            abort(400)

        mts = MembershipTokenSerializer()
        inv_id = community_member.invitation_id

        if community_member.request.is_invite:
            if not token or not mts.validate_token(
                    token, expected_value=inv_id):
                abort(404)

        if action == 'accept':
            community_member.status = CommunityMemberStatus.ACCEPTED
            if community_member.request.is_invite:
                community_member.user_id = request_user.id
            if community_member.invitation_id:
                community_member.invitation_id = None
            if role:
                if request_user in community.members.filter(
                        role=CommunityMemberRole.ADMIN,
                        status=CommunityMemberStatus.ACCEPTED):
                    community_member.role = CommunityMemberRole.from_str(role)
                else:
                    abort(400)
            community_member.request.close_request()
        elif action == 'reject':
            #TODO check if the invitation_id is properly cleaned up
            if community_member.invitation_id:
                community_member.invitation_id = None
            community_member.status = CommunityMemberStatus.REJECTED
            community_member.request.close_request()
        if comment:
            community_member.request.add_comment(request_user.id, comment)
        community_member.request.commit()
        db.session.commit()
        # TODO response?
        return jsonify(community_member.as_dict()), 200


api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>/members'.format(pid=comid_url_converter),
    view_func=ListResource.as_view('community_members_api'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>/members/<any({status}):status>'.format(
        pid=comid_url_converter,
        status=','.join(['accepted', 'rejected', 'pending'])),
    view_func=ListMembersResource.as_view('community_members_list_api'),
)


api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/members/requests/<membership_id>/<any({actions}):action>'.format(
        pid=comid_url_converter,
        actions=','.join(['accept', 'reject', 'comment'])),
    view_func=MembershipRequestHandlingResource.as_view(
        'community_requests_handling_api'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>/members/requests/<membership_id>'.format(
        pid=comid_url_converter),
    view_func=MembershipRequestResource.as_view(
        'community_requests_api'),
)


ui_blueprint = Blueprint(
    'invenio_communities_members',
    __name__,
    template_folder='../templates',
)


@ui_blueprint.route(
    '/communities/<{}:pid_value>/members'.format(comid_url_converter))
@pass_community
@login_required
def members(comid=None, community=None):
    """Members of a community."""
    community_member_obj = community.members.get_user_membership(
        int(current_user.get_id()),
        status=CommunityMemberStatus.ACCEPTED
    )
    if community_member_obj is not None:
        community_member = community_member_obj.as_dict()
        pending_records = \
            len(community.members.filter(status=CommunityMemberStatus.PENDING))

    else:
        community_member = {}
        pending_records = {}
    #TODO pending records should also check permissions
    #  before exposing this information
    return render_template(
        'invenio_communities/members.html',
        community=community,
        comid=comid,
        pending_records=pending_records,
        community_member=community_member
    )


@ui_blueprint.route(
    '/communities/<{}:pid_value>/members/requests/<membership_id>'.format(
        comid_url_converter))
@pass_community
@pass_membership
@login_required
def requests(
        comid=None, community=None, community_member=None):
    """Requests of communities."""
    token = request.args.get('token', '')
    if not community_member.request.is_invite:
        abort(404)
    mts = MembershipTokenSerializer()
    inv_id = community_member.invitation_id

    if not token or not mts.validate_token(
            token, expected_value=inv_id):
        abort(404)
    return render_template(
        'invenio_communities/request.html',
        comid=comid,
        community=community,
        membership=community_member.as_dict(include_requests=True),
        token=token
    )



from __future__ import absolute_import, print_function

import binascii
import os
from base64 import urlsafe_b64encode
from datetime import datetime

from flask import current_app
from itsdangerous import BadData, JSONWebSignatureSerializer, SignatureExpired

SUPPORTED_DIGEST_ALGORITHMS = ('HS256', 'HS512')


class TokenMixin(object):
    """Mix-in class for token serializers."""

    def create_token(self, obj_id, extra_data={}):
        """Create a token referencing the object id with extra data.
        Note random data is added to ensure that no two tokens are identical.
        """
        return self.dumps(
            dict(
                id=obj_id,
                data=extra_data,
                rnd=binascii.hexlify(os.urandom(4)).decode('utf-8')
            )
        )

    def validate_token(self, token, expected_value=None):
        """Validate secret link token.
        :param token: Token value.
        :param expected_value: A dictionary of key/values that must be present
            in the data part of the token (i.e. included via ``extra_data`` in
            ``create_token``).
        """
        try:
            # Load token and remove random data.
            data = self.load_token(token)

            # Compare expected data with data in token.
            if expected_value:
                if expected_value != data["id"]:
                    return None
            return data
        except BadData:
            return None

    def load_token(self, token, force=False):
        """Load data in a token.
        :param token: Token to load.
        :param force: Load token data even if signature expired.
                      Default: False.
        """
        try:
            data = self.loads(token)
        except SignatureExpired as e:
            if not force:
                raise
            data = e.payload

        del data["rnd"]
        return data


class MembershipTokenSerializer(JSONWebSignatureSerializer, TokenMixin):
    """Serializer for email confirmation link tokens.
    Depends upon the secrecy of ``SECRET_KEY``. Tokens expire after a specific
    time (defaults to ``ACCESSREQUESTS_CONFIRMLINK_EXPIRES_IN``). The
    access request id as well as the email address is stored in the token
    together with a random bit to ensure all tokens are unique.
    """

    def __init__(self, expires_in=None, **kwargs):
        """Initialize underlying TimedJSONWebSignatureSerializer."""
        super(MembershipTokenSerializer, self).__init__(
            current_app.config['SECRET_KEY'],
            salt='membershiprequests-email',
            **kwargs
        )

    @classmethod
    def compat_validate_token(cls, *args, **kwargs):
        """Multiple algorithm-compatible token validation."""
        data = None
        for algorithm in SUPPORTED_DIGEST_ALGORITHMS:
            data = cls(algorithm_name=algorithm).validate_token(
                *args, **kwargs)
            if not data:  # move to next algorithm
                continue
        return data



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community record permissions."""

from flask_principal import UserNeed
from invenio_access.permissions import any_user
from invenio_records_permissions.generators import Generator
from invenio_records_permissions.policies import BasePermissionPolicy

from invenio_communities.permissions import CommunityOwner


class RecordOwners(Generator):
    """Allows record owners."""

    def needs(self, inclusion_request=None, record=None, **kwargs):
        """Enabling Needs."""
        record = record or inclusion_request.record
        return [UserNeed(owner) for owner in record.get('_owners', [])]


class InclusionRequestHandlers(Generator):

    def needs(self, inclusion_request=None, community=None, **kwargs):
        """Skip check at this stage if the membership is an invitation."""
        if inclusion_request['state'] == inclusion_request.STATES['CLOSED']:
            return []
        if inclusion_request.is_request:
            return CommunityOwner().needs(community=community)
        else:
            return RecordOwners().needs(record=inclusion_request.record)


class CommunityRecordPermissionPolicy(BasePermissionPolicy):

    can_list_community_inclusions = [CommunityOwner()]
    can_create_community_inclusion = [RecordOwners(), CommunityOwner()]
    can_get_community_inclusion = [RecordOwners(), CommunityOwner()]
    can_delete_community_inclusion = [RecordOwners(), CommunityOwner()]
    can_handle_community_inclusion = [InclusionRequestHandlers()]
    can_comment_community_inclusion = [RecordOwners(), CommunityOwner()]


def is_permitted_action(action, **kwargs):
    """Helper function to apply a permission check."""
    return CommunityRecordPermissionPolicy(action=action, **kwargs)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Blueprint definitions for records integration."""

import uuid
from functools import wraps

from flask import Blueprint, abort, jsonify, render_template, request
from flask.views import MethodView
from flask_login import current_user, login_required
from invenio_db import db
from invenio_indexer.api import RecordIndexer
from invenio_pidstore.errors import PIDDoesNotExistError
from sqlalchemy.exc import SQLAlchemyError
from webargs import fields

from invenio_communities.records.api import CommunityInclusionRequest, \
    CommunityRecord, CommunityRecordsCollection, Record, \
    RecordCommunitiesCollection

from ..utils import comid_url_converter
from ..views import pass_community, use_kwargs
from .errors import CommunityRecordAlreadyExists
from .permissions import CommunityRecordPermissionPolicy, is_permitted_action

#
# UI views
#
ui_blueprint = Blueprint(
    'invenio_communities_records',
    __name__,
    template_folder='../templates',
)


def pass_record(func=None):
    """Decorator to retrieve community membership."""
    @wraps(func)
    def inner(*args, record_pid=None, **kwargs):
        try:
            record_pid, record = CommunityRecord.record_cls.resolve(record_pid)
        except PIDDoesNotExistError:
            abort(404)
        return func(*args, record_pid=record_pid, record=record, **kwargs)
    return inner


def pass_request(func=None):
    """Decorator to retrieve community membership."""
    @wraps(func)
    def inner(*args, **kwargs):
        del kwargs['request_id']
        request_id = request.view_args['request_id']
        inclusion_request = CommunityInclusionRequest.get_record(request_id)
        if inclusion_request is None:
            abort(404)
        return func(*args, inclusion_request=inclusion_request, **kwargs)
    return inner


def community_permission(
        action=None,
        error_message='You are missing the permissions to '
                      'access this information.',
        error_code=403):
    """Wrapper to apply a permission check on a view."""
    def wrapper_function(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not CommunityRecordPermissionPolicy(action=action, **kwargs):
                abort(error_code, error_message)
            return func(*args, **kwargs)
        return wrapper
    return wrapper_function


@ui_blueprint.route(
    '/communities/<{}:pid_value>/curate'.format(comid_url_converter))
@pass_community
@login_required
@community_permission('list_community_inclusions')
def curation(comid=None, community=None):
    """Curate page for community record inclusions."""
    pending_records = \
        len(CommunityRecordsCollection(community).filter({'status': 'P'}))

    return render_template(
        'invenio_communities/curation.html',
        community=community,
        comid=comid,
        pending_records=pending_records
    )


#
# REST API views
#
api_blueprint = Blueprint(
    'invenio_communities_records',
    __name__,
    template_folder='../templates',
)


class ListResource(MethodView):
    """Resource for listing and creating inclusion requests."""

    post_args = {
        'record_pid': fields.String(
            location='json',
            required=True,
        ),
        'message': fields.String(
            location='json',
            required=False,
        ),
        'auto_accept': fields.Boolean(
            location='json',
            required=False,
        )
    }

    @pass_community
    @login_required
    @community_permission('list_community_inclusions')
    def get(self, comid=None, community=None):
        """List the community records requests."""
        json_response = community.records.as_dict()
        return jsonify(json_response), 200

    @use_kwargs(post_args)
    @pass_community
    @pass_record
    @login_required
    @community_permission('create_community_inclusion')
    def post(self, comid=None, community=None, record=None, record_pid=None,
             message=None, auto_accept=False, **kwargs):
        """Join a community or invite a user record to it."""
        #
        # View - context parsing/building
        #
        user = current_user

        #
        # Controller
        #
        # TODO: implement "__contains__" in members collection API
        # admin_ids = \
        #     [admin.user.id for admin in CommunityMember.get_admins(
        #         community.id)]
        # if int(current_user.get_id()) not in admin_ids:
        #     abort(404)

        # TODO: Consider also
        # "CommunityInclusionRequest.community_record_cls.record_cls.resolve"
        request = CommunityInclusionRequest.create(user, id_=uuid.uuid4())
        if message:
            request.add_comment(user, message)
        try:
            com_rec = CommunityRecord.create(
                record, record_pid, community, request)
        except CommunityRecordAlreadyExists:
            abort(400, 'Community-record relationship already exists.')
        com_rec.status = com_rec.Status.PENDING

        # TODO: Check in community members instead of just creator
        # if CommunityMembersAPI.has_member(community, user):
        if user.id in record['_owners']:
            # if user is also a community owner and auto-accepted:
            if auto_accept and community['created_by'] == user.id:
                # TODO: define what the routing_key value should be
                # request.routing_key = ''
                # TODO: implement state as a "Request.state" property?
                request['state'] = request.State['CLOSED']
                com_rec.status = com_rec.Status.ACCEPTED
                request['inclusion_type'] = 'auto-accepted'
            else:
                # TODO: define what the routing_key value should be
                # TODO: don't use f-strings
                # request.routing_key = f'record:{record_pid.pid_value}:owners'
                request['inclusion_type'] = 'request'
        else:
            # TODO: define what the routing_key value should be
            # request.routing_key = \
            #     f'community:{community.pid.pid_value}:curators'
            request['inclusion_type'] = 'invite'
        db.session.commit()

        # Notify request owners and receivers
        # TODO: implement mail sending (e.g. via "routing_key")
        # send_request_emails(request)

        # Index record with new inclusion request info
        # TODO: Implement indexer receiver to include community info in record
        RecordIndexer().index_by_id(record.id)

        json_response = request.as_dict()
        json_response['links'] = request.dump_links(request, comid.pid_value)

        #
        # View - serialize response
        #
        # TODO: Use marshmallow to serialize
        return jsonify(json_response), 201


class ItemResource(MethodView):
    """Community inclusion request item endpoint."""

    @pass_community
    @pass_request
    @login_required
    @community_permission('get_community_inclusion')
    def get(self, comid=None, community=None, inclusion_request=None):
        """Get the inclusion request."""
        # TODO: check if user in community or record owner
        json_response = inclusion_request.as_dict()
        return jsonify(json_response), 200

    @pass_community
    @pass_request
    @login_required
    @community_permission('delete_community_inclusion')
    def delete(self, comid=None, community=None, inclusion_request=None):
        """Delete the inclusion request."""
        # TODO: check if user in community or record owner
        community_record = request.community_record
        community_record.delete()
        request.delete()
        db.session.commit()
        return 204


# TODO: move things around
class ItemActionsResource(MethodView):
    """Perform actions on inclusion requests."""

    post_args = {
        'message': fields.String(
            location='json',
            required=False,
        )
    }

    @use_kwargs(post_args)
    @pass_community
    @pass_request
    @login_required
    # TODO: Swap `request_id` with `CommunityRecord.id` (instead of `Request.id`)
    def post(self, comid=None, community=None, inclusion_request=None,
             action=None, message=None, **kwargs):
        """Handle a community record request."""
        #
        # View - context parsing/building
        #
        community_record = inclusion_request.community_record
        if action == 'comment':
            if not is_permitted_action(
                    'comment_community_inclusion',
                    inclusion_request=inclusion_request, community=community):
                abort(404)
            if not message:
                # TODO: add message missing
                abort(400, 'Missing comment message.')
        elif not is_permitted_action(
                'handle_community_inclusion',
                inclusion_request=inclusion_request, community=community):
            abort(404)
        elif action == 'accept':
            community_record.status = community_record.Status.ACCEPTED
            # TODO: figure out "routing_key"
            # inclusion_request.routing_key = ''
            # TODO: use ".state" property setter/getter
            inclusion_request['state'] = inclusion_request.STATES['CLOSED']
        elif action == 'reject':
            community_record.status = community_record.Status.REJECTED
            # TODO: figure out "routing_key"
            # inclusion_request.routing_key = ''
            inclusion_request['state'] = inclusion_request.STATES['CLOSED']
        else:
            # TODO: no appropriate action error
            abort(400, 'Invalid action')
        if message:
            inclusion_request.add_comment(current_user, message)
        db.session.commit()

        # Notify request owners and receivers
        # TODO: implement mail sending
        # send_request_emails(request)

        # Index record with new inclusion request info
        # TODO: Implement indexer receiver to include community info in record
        # TODO: check if it's always needed to index the record
        RecordIndexer().index_by_id(community_record.record.id)

        #
        # View - serialize response
        #
        # TODO: Use marshmallow to serialize
        return jsonify(community_record.as_dict()), 201


api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion'.format(pid=comid_url_converter),
    view_func=ListResource.as_view('requests_list'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion/<uuid:request_id>'.format(pid=comid_url_converter),
    view_func=ItemResource.as_view('requests_item'),
)

api_blueprint.add_url_rule(
    '/communities/<{pid}:pid_value>'
    '/requests/inclusion/<uuid:request_id>'
    '/<any({actions}):action>'.format(
        pid=comid_url_converter,
        actions=','.join(['accept', 'reject', 'comment'])),
    view_func=ItemActionsResource.as_view('requests_item_actions'),
)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2017-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.
"""Identity changed loaders."""

from flask_security import current_user

from invenio_communities.members.models import CommunityMember, \
    CommunityMemberStatus

from .members.permissions import CommunityMemberNeed


def load_permissions_on_identity_loaded(sender, identity):
    """Add system roles "Needs" to users' identities.
    Every user gets the **any_user** Need.
    Authenticated users get in addition the **authenticated_user** Need.
    """
    if current_user.is_authenticated:
        CommunityMember.query.filter_by(
            status=CommunityMemberStatus.ACCEPTED,
            user_id=int(current_user.get_id()))
        for membership in CommunityMember.query.filter_by(
                user_id=int(current_user.get_id()),
                status=CommunityMemberStatus.ACCEPTED):
            if membership.status == CommunityMemberStatus.ACCEPTED:
                identity.provides.add(CommunityMemberNeed(
                    membership.community_pid.id,
                    membership.role.name.lower()
                ))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community database models."""

from __future__ import absolute_import, print_function

from enum import Enum

from flask_babelex import gettext
from invenio_accounts.models import User
from invenio_db import db
from invenio_pidstore.models import PersistentIdentifier
from invenio_records.models import RecordMetadataBase
from speaklater import make_lazy_gettext
from sqlalchemy.exc import IntegrityError
from sqlalchemy_utils.types import ChoiceType, UUIDType

from invenio_communities.requests.models import Request

from .errors import CommunityMemberAlreadyExists

_ = make_lazy_gettext(lambda: gettext)

COMMUNITY_MEMBER_TITLES = {
    'MEMBER': _('Member'),
    'ADMIN': _('Admin'),
    'CURATOR': _('Curator')
}


class CommunityMemberRole(Enum):
    """Constants for possible roles of any given Community member."""
    # TODO Make roles configurable.
    MEMBER = 'M'
    """Member of the community."""

    ADMIN = 'A'
    """Admin of the community."""

    CURATOR = 'C'
    """Curator of the community."""

    @property
    def title(self):
        """Return human readable title."""
        return COMMUNITY_MEMBER_TITLES[self.name]

    @classmethod
    def from_str(cls, val):
        return getattr(cls, val.upper())


COMMUNITY_MEMBER_STATUS = {
    'PENDING': _('Pending'),
    'ACCEPTED': _('Accepted'),
    'REJECTED': _('Rejected'),
    'BLOCKED': _('Blocked')
}


class CommunityMemberStatus(Enum):
    """Community-member relationship status."""

    PENDING = 'P'
    ACCEPTED = 'A'
    REJECTED = 'R'
    BLOCKED = 'B'

    @property
    def title(self):
        """Return human readable title."""
        return COMMUNITY_MEMBER_STATUS[self.name]

    @classmethod
    def from_str(cls, val):
        return getattr(cls, val.upper())


class CommunityMember(db.Model, RecordMetadataBase):
    """Represent a community member role."""

    __tablename__ = 'communities_members'
    __versioned__ = {'versioning': False}
    __table_args__ = (
        db.Index(
            'uidx_community_pid_id_user_id_invitation_id',
            # TODO: Check if this combination would be enough (since user_id can be NULL)
            'community_pid_id', 'user_id',
            # 'community_pid_id', 'user_id', 'invitation_id',
            unique=True,
        ),
        db.Index(
            'uidx_community_pid_id_invitation_id',
            'community_pid_id', 'invitation_id',
            unique=True,
        ),
        {'extend_existing': True},
    )

    community_pid_id = db.Column(
        db.Integer,
        db.ForeignKey(PersistentIdentifier.id),
        nullable=False,
    )
    """Community PID ID."""

    user_id = db.Column(
        db.Integer,
        db.ForeignKey(User.id),
    )
    """User ID."""

    request_id = db.Column(
        UUIDType,
        db.ForeignKey(Request.id),
        # TODO: should we also allow CommunityMembers without a request?
        nullable=True,
    )
    """Request ID."""

    invitation_id = db.Column(
        db.String(255),
        nullable=True,
    )
    """Invitation ID."""

    status = db.Column(
        ChoiceType(CommunityMemberStatus, impl=db.CHAR(1)),
        nullable=False,
        default=CommunityMemberStatus.PENDING,
    )

    role = db.Column(
        ChoiceType(CommunityMemberRole, impl=db.CHAR(1)), nullable=False)

    user = db.relationship(User, backref='communities')

    community_pid = db.relationship(PersistentIdentifier)

    request = db.relationship(Request)

    @classmethod
    def create(cls, community_pid_id, request_id=None, role='M', status='P',
               user_id=None, invitation_id=None, json=None):
        """Create Community Membership Role."""
        try:
            with db.session.begin_nested():
                obj = cls(
                    community_pid_id=community_pid_id,
                    user_id=user_id,
                    request_id=request_id,
                    invitation_id=invitation_id,
                    role=role,
                    status=status,
                    json=json,
                )
                db.session.add(obj)
        except IntegrityError as ex:
            # TODO: Differentiate between the two possible unique constraint errors
            # that the DB might raise
            raise CommunityMemberAlreadyExists(
                pid_id=community_pid_id,
                user_id=user_id,
                invitation_id=invitation_id,
            )
        return obj

    @classmethod
    def delete(cls, community_member):
        """Delete community member relationship."""
        with db.session.begin_nested():
            db.session.delete(community_member)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community permissions."""

from flask_principal import UserNeed
from flask_security import current_user
from invenio_access.permissions import any_user
from invenio_records_permissions.generators import Generator
from invenio_records_permissions.policies import BasePermissionPolicy

from invenio_communities.members.permissions import CommunityMember


class CommunityVisibilityPolicy(Generator):

    def needs(self, community=None, **kwargs):
        if community['visibility'] == 'public':
            return [any_user]
        else:
            return [CommunityMember(['admin', 'curator', 'member'])]


class CommunityOwner(Generator):
    """Returns community owner need."""
    def needs(self, community=None, **kwargs):
        """."""
        return [UserNeed(community['created_by'])]


class CommunityPermissionPolicy(BasePermissionPolicy):

    can_read_community = [
        CommunityVisibilityPolicy(),
    ]

    can_update_community = [
        CommunityMember(['admin', 'curator']),
    ]
    can_delete_community = [
        CommunityMember(['admin']),
    ]



def allow_logged_in(*args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        return current_user.is_authenticated
    return type('AllowCommunityOwner', (), {'can': can})()


def can_read_community(record, *args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        return CommunityPermissionPolicy(
            action='read_community',
            comid=record.pid,
            community=record,
        ).can()
    return type('AllowCommunityOwner', (), {'can': can})()


def can_update_community(record, *args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        return CommunityPermissionPolicy(
            action='update_community',
            comid=record.pid,
            community=record,
        ).can()
    return type('AllowCommunityOwner', (), {'can': can})()


def can_delete_community(record, *args, **kwargs):
    """Permission that checks if the community owner is making the request."""
    def can(self):
        """Check that the current user is the community's owner."""
        return CommunityPermissionPolicy(
            action='delete_community',
            comid=record.pid,
            community=record,
        ).can()
    return type('AllowCommunityOwner', (), {'can': can})()



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Invenio communities extension."""

from __future__ import absolute_import, print_function

from flask_principal import identity_loaded
from invenio_base.utils import obj_or_import_string
from invenio_indexer.signals import before_record_index
from werkzeug.utils import cached_property

from . import config
from .permission_loaders import load_permissions_on_identity_loaded
from .signals import community_created
from .utils import LazyPIDConverter, obj_or_import_string, set_default_admin


class InvenioCommunities(object):
    """Invenio extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    @cached_property
    def community_cls(self):
        """Base community API class."""
        # TODO: Refactor
        # def default_class_factory():
        #     from .api import CommunityBase
        #     from .records.api import CommunityRecordsMixin
        #     from .members.api import CommunityMembersMixin
        #     return type(
        #         'Community',
        #         (CommunityBase, CommunityRecordsMixin, CommunityMembersMixin),
        #         {},
        #     )
        # return self.app.config['COMMUNITY_CLS'] or default_class_factory()

        from .api import CommunityBase
        from .members.api import CommunityMembersMixin
        from .records.api import CommunityRecordsMixin
        return type(
            'Community',
            (CommunityBase, CommunityRecordsMixin, CommunityMembersMixin),
            {},
        )

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        app.extensions['invenio-communities'] = self

        # TODO: Remove when merged in invenio-records-rest
        app.url_map.converters['lazy_pid'] = LazyPIDConverter

        @app.context_processor
        def record_context_processors():
            """Jinja context processors for communities record integration."""
            def record_communities(record):
                from invenio_communities.records.api import Record, \
                    RecordCommunitiesCollection
                return RecordCommunitiesCollection(Record(record, model=record.model))
            return dict(record_communities=record_communities)

        self._register_signals(app)

        identity_loaded.connect_via(app)(
            load_permissions_on_identity_loaded
        )

    def init_config(self, app):
        """Initialize configuration.

        Override configuration variables with the values in this package.
        """
        for k in dir(config):
            if k.startswith('COMMUNITIES_'):
                if k == 'COMMUNITIES_REST_ENDPOINTS':
                    # Make sure of registration process.
                    app.config.setdefault('RECORDS_REST_ENDPOINTS', {})
                    app.config['RECORDS_REST_ENDPOINTS'].update(getattr(
                        config, k))
                app.config.setdefault(k, getattr(config, k))
                if k == 'COMMUNITIES_REST_FACETS':
                    app.config['RECORDS_REST_FACETS'].update(
                        getattr(config, k))
                if k == 'COMMUNITIES_REST_SORT_OPTIONS':
                    # TODO Might be overriden depending on which package is
                    # initialised first
                    app.config.setdefault('RECORDS_REST_SORT_OPTIONS', {})
                    app.config['RECORDS_REST_SORT_OPTIONS'].update(
                        getattr(config, k))
                if k == 'COMMUNITIES_REST_DEFAULT_SORT':
                    # TODO Might be overriden depending on which package is
                    # initialised first
                    app.config.setdefault('RECORDS_REST_DEFAULT_SORT', {})
                    app.config['RECORDS_REST_DEFAULT_SORT'].update(
                        getattr(config, k))

        app.config.setdefault(
            'SUPPORT_EMAIL', getattr(config, 'SUPPORT_EMAIL'))

    def _register_signals(self, app):
        """Register signals."""
        # TODO: Make configurable or move into separate extension in ".records"
        from invenio_communities.records.indexer import record_indexer_receiver
        before_record_index.dynamic_connect(
            obj_or_import_string(
                app.config.get('COMMUNITIES_INDEXER_RECEIVER'),
                record_indexer_receiver),
            sender=app,
            weak=False,
            index=app.config.get(
                'COMMUNITIES_RECORD_INDEX', 'records-record-v1.0.0'),
        )
        # TODO: Make it possible to enable/disable the members feature via config
        community_created.connect(set_default_admin, weak=False)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration.

See https://pytest-invenio.readthedocs.io/ for documentation on which test
fixtures are available.
"""
import pytest
from invenio_accounts.testutils import create_test_user, login_user_via_session
from invenio_app.factory import create_api


@pytest.fixture(scope='module')
def create_app(instance_path):
    """Application factory fixture."""
    return create_api


@pytest.fixture(scope='module')
def basic_user(app):
    yield create_test_user('test@inveniosoftware.org')


@pytest.fixture(scope='module')
def users(app):
    yield [create_test_user('user{}@inveniosoftware.org'.format(i)) for i in range(3)]



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Default configuration."""

from __future__ import absolute_import, print_function

from invenio_records_rest.facets import terms_filter
from invenio_records_rest.utils import allow_all

from .permissions import allow_logged_in, can_delete_community, \
    can_read_community, can_update_community
from .utils import comid_url_converter

#: Records REST API endpoints.
COMMUNITIES_REST_ENDPOINTS = dict(
    comid=dict(
        pid_type='comid',
        pid_minter='comid',
        pid_fetcher='comid',
        list_route='/communities/',
        item_route='/communities/<{}:pid_value>'.format(comid_url_converter),
        search_index='communities',
        record_class='invenio_communities.api:Community',
        record_serializers={
            'application/json': (
                'invenio_communities.serializers.json_v1_response'),
        },
        search_serializers={
            'application/json': (
                'invenio_communities.serializers:json_v1_search'),
        },
        record_loaders={
            'application/json': ('invenio_communities.loaders:json_v1'),
        },
        indexer_class='invenio_communities.indexer:CommunityIndexer',
        default_media_type='application/json',
        read_permission_factory_imp=can_read_community,
        create_permission_factory_imp=allow_logged_in,
        list_permission_factory_imp=allow_all,
        update_permission_factory_imp=can_update_community,
        delete_permission_factory_imp=can_delete_community,
    ),
)

COMMUNITIES_REST_FACETS = dict(
    communities=dict(
        aggs=dict(
            type=dict(
                terms=dict(field="type"),
            ),
            domain=dict(
                terms=dict(field="domains"),
            )
        ),
        post_filters=dict(
            type=terms_filter('type'),
            domain=terms_filter('domains'),
        )
    )
)

COMMUNITIES_REST_SORT_OPTIONS = dict(
    communities=dict(
        bestmatch=dict(
            title='Best match',
            fields=['_score'],
            default_order='desc',
            order=1,
        ),
        mostrecent=dict(
            title='Most recent',
            fields=['-_created'],
            default_order='asc',
            order=2,
        ),
    )
)

COMMUNITIES_REST_DEFAULT_SORT = dict(
    communities=dict(
        query='bestmatch',
        noquery='mostrecent',
    )
)

COMMUNITIES_MEMBERSHIP_REQUESTS_CONFIRMLINK_EXPIRES_IN = 1000000

SUPPORT_EMAIL = 'test@email.org'

COMMUNITIES_RECORD_INDEX = 'records-record-v1.0.0'

# TODO: Use vocabulary-based model
COMMUNITIES_DOMAINS = [
    {'text': 'Agricultural and Veterinary Sciences', 'value': 'agricultural_and_veterinary_sciences'},
    {'text': 'Biological Sciences', 'value': 'biological_sciences'},
    {'text': 'Built Environment and Design', 'value': 'built_environment_and_design'},
    {'text': 'Chemical Sciences', 'value': 'chemical_sciences'},
    {'text': 'Commerce, Management, Tourism and Services', 'value': 'commerce_management_tourism_and_services'},
    {'text': 'Earth Sciences', 'value': 'earth_sciences'},
    {'text': 'Environmental Sciences', 'value': 'environmental_sciences'},
    {'text': 'Economics', 'value': 'economics'},
    {'text': 'Education', 'value': 'education'},
    {'text': 'Engineering', 'value': 'engineering'},
    {'text': 'Technology', 'value': 'technology'},
    {'text': 'History and Archaeology', 'value': 'history_and_archaeology'},
    {'text': 'Information and Computing Sciences', 'value': 'information_and_computing_sciences'},
    {'text': 'Language, Communication and Culture', 'value': 'language_communication_and_culture'},
    {'text': 'Law and Legal Studies', 'value': 'law_and_legal_studies'},
    {'text': 'Mathematical Sciences', 'value': 'mathematical_sciences'},
    {'text': 'Medical and Health Sciences', 'value': 'medical_and_health_sciences'},
    {'text': 'Philosophy and Religious Studies', 'value': 'philosophy_and_religious_studies'},
    {'text': 'Physical Sciences', 'value': 'physical_sciences'},
    {'text': 'Psychology and Cognitive Sciences', 'value': 'psychology_and_cognitive_sciences'},
    {'text': 'Studies in Creative Arts and Writing', 'value': 'studies_in_creative_arts_and_writing'},
    {'text': 'Studies in Human Society', 'value': 'studies_in_human_society'},
]

# Overwite to change the default indexer for community records
# COMMUNITIES_INDEXER_RECEIVER = None

# Overwrite to change generated URL for request management
# COMMUNITIES_INVITATION_LINK_URL = None

# Overwrite to change available community member roles
# COMMUNITY_MEMBER_ROLES = None

# Overwrite to change available community member status options
# COMMUNITY_MEMBER_STATUS = None



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2020 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community module tests."""

from flask import url_for
from invenio_accounts.testutils import create_test_user, login_user_via_session
from invenio_search import current_search


def assert_error_resp(res, expected_errors, expected_status_code=400):
    """Assert errors in a client response."""
    assert res.status_code == expected_status_code
    payload = res.json
    errors = payload.get('errors', [])
    for field, msg in expected_errors:
        if not field:  # top-level "message" error
            assert msg in payload['message'].lower()
            continue
        assert any(
            e['field'] == field and msg in e['message'].lower()
            for e in errors), payload


def create_test_community(client, user, data):
    """Create test community."""
    login_user_via_session(client, user=user)
    list_url = url_for('invenio_records_rest.comid_list')
    resp = client.post(list_url, json=data)
    assert resp.status_code == 201
    return resp


def logout_user_session(client):
    """Logout a user from the Flask test client session."""
    with client.session_transaction() as sess:
        del sess['user_id']


def test_simple_flow(db, es_clear, client, users):
    """Test community minimal CRUD operations."""
    community_owner, _, _ = users
    list_url = url_for('invenio_records_rest.comid_list')
    login_user_via_session(client, user=community_owner)

    # Create
    resp = client.post(list_url, json={
        'id': 'minimal_comm',
        'title': 'Minimal community',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 201
    assert resp.json['metadata']['id'] == 'minimal_comm'
    assert resp.json['metadata']['title'] == 'Minimal community'
    assert resp.json['metadata']['type'] == 'organization'
    assert resp.json['links']['self'] == 'http://localhost/communities/minimal_comm'

    item_url = resp.json['links']['self']

    # Read
    resp = client.get(item_url)
    assert resp.status_code == 200
    assert resp.json['metadata']['id'] == 'minimal_comm'
    assert resp.json['metadata']['title'] == 'Minimal community'
    assert resp.json['metadata']['type'] == 'organization'
    assert resp.json['links']['self'] == 'http://localhost/communities/minimal_comm'

    # Update
    resp = client.put(item_url, json={
        'id': 'biosyslit',
        'title': 'Owner title',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 200
    assert resp.json['metadata']['title'] == 'Owner title'

    # Delete
    resp = client.delete(item_url)
    assert resp.status_code == 204

    logout_user_session(client)


def test_community_permissions(db, es_clear, client, users):
    """Test community CRUD operations."""
    # TODO: test for anonymous and regular user => not for owners
    community_owner, regular_user, _ = users
    list_url = url_for('invenio_records_rest.comid_list')

    # create demo community
    data = {
        'id': 'test_comm',
        'title': 'Testing community',
        'type': 'organization',
        'visibility': 'public',
    }
    resp = create_test_community(client, community_owner, data)
    item_url_owner = resp.json['links']['self']
    current_search.flush_and_refresh('communities')

    # ANONYMOUS
    logout_user_session(client)
    # create
    resp = client.post(list_url, json={
        'id': 'anonymous_comm',
        'title': 'Anonymous community',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 401

    # read
    resp = client.get(list_url)
    assert resp.status_code == 200
    assert resp.json['hits']['total'] == 1

    resp = client.get(item_url_owner)
    assert resp.status_code == 200
    assert resp.json['metadata']['title'] == 'Testing community'
    assert resp.json['metadata']['type'] == 'organization'
    assert resp.json['metadata']['id'] == 'test_comm'
    assert resp.json['links']['self'] == 'http://localhost/communities/test_comm'


    # update
    resp = client.put(item_url_owner, json={
        'id': 'biosyslit',
        'title': 'Owner title',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 401

    # delete
    resp = client.delete(item_url_owner)
    assert resp.status_code == 401

    # REGULAR USER
    login_user_via_session(client, user=regular_user)

    # create a community
    resp = client.post(list_url, json={
        'id': 'regular_comm',
        'title': 'Regular community',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 201
    assert resp.json['metadata']['id'] == 'regular_comm'
    assert resp.json['metadata']['title'] == 'Regular community'
    assert resp.json['metadata']['type'] == 'organization'
    assert resp.json['links']['self'] == 'http://localhost/communities/regular_comm'

    item_url_regular = resp.json['links']['self']
    current_search.flush_and_refresh('communities')

    # read
    resp = client.get(list_url)
    assert resp.status_code == 200
    assert resp.json['hits']['total'] == 2

    resp = client.get(item_url_regular)
    assert resp.status_code == 200
    assert resp.json['metadata']['id'] == 'regular_comm'
    assert resp.json['metadata']['title'] == 'Regular community'
    assert resp.json['metadata']['type'] == 'organization'
    assert resp.json['links']['self'] == 'http://localhost/communities/regular_comm'

    # update another community
    resp = client.put(item_url_owner, json={
        'id': 'biosyslit',
        'title': 'Regular title',
        'type': 'organization',
        'visibility': 'public',
    })

    assert resp.status_code == 403

    # Delete community
    resp = client.delete(item_url_owner)
    assert resp.status_code == 403

    # Logout
    logout_user_session(client)


def test_community_validation(db, es_clear, client, users):
    """Test community validation."""
    community_owner, regular_user, _ = users
    list_url = url_for('invenio_records_rest.comid_list')
    login_user_via_session(client, user=community_owner)

    # missing field
    resp = client.post(list_url, json={
        'id': 'biosyslit_val',
        'title': 'BLRval',
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert_error_resp(resp, (
        ('type', 'required field'),
    ))

    # wrong type field
    resp = client.post(list_url, json={
        'id': 'biosyslit_val',
        'title': 'BLRval',
        'type': 'wrong type'
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert_error_resp(resp, (
        ('type', 'must be one of'),
    ))

    # id already created
    resp = client.post(list_url, json={
        'id': 'comm_id',
        'title': 'Original title',
        'type': 'project'
    })

    assert resp.status_code == 201

    resp = client.post(list_url, json={
        'id': 'comm_id',
        'title': 'Duplicate title',
        'type': 'project'
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert_error_resp(resp, (
        ('id', 'is already assigned to a community'),
    ))

    # wrong domain
    resp = client.post(list_url, json={
        'id': 'domain_id',
        'title': 'Domain title',
        'type': 'project',
        'domains': ['wrong']
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert 'Must be one of' in resp.json['errors'][0]['message']

    # wrong domain formatting
    resp = client.post(list_url, json={
        'id': 'domain_id',
        'title': 'Domain title',
        'type': 'project',
        'domains': 'engineering'
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert_error_resp(resp, (
        ('domains', 'not a valid list'),
    ))

    # wrong field
    resp = client.post(list_url, json={
        'id': 'domain_id',
        'title': 'Domain title',
        'type': 'project',
        'field': 'wrong'
    })

    assert resp.status_code == 400
    assert resp.json['message'] == 'Validation error.'
    assert_error_resp(resp, (
        ('field', 'unknown field'),
    ))

    # test all fields
    resp = client.post(list_url, json={
        'id': 'full_comm',
        'title': 'All fields',
        'type': 'event',
        'curation_policy': 'Policy field',
        'page': 'Page field',
        'website': 'http://www.fullfield.com',
        'funding': ['OpenAire'],
        'domains': ['engineering', 'technology'],
        'verified': True,
        'visibility': 'public',
        'member_policy': 'open',
        'record_policy': 'open',
        'archived': 'False',
    })

    assert resp.status_code == 201


def test_community_search(db, es_clear, client, users):
    """Test community search."""


    community_owner, _ , _ = users
    list_url = url_for('invenio_records_rest.comid_list')
    login_user_via_session(client, user=community_owner)

    # creating demos communities
    resp = client.post(list_url, json={
        'id': 'regular_comm',
        'title': 'Regular community',
        'type': 'topic',
        'domains': ['engineering']
    })
    assert resp.status_code == 201

    resp = client.post(list_url, json={
        'id': 'biosyslit',
        'title': 'BLR',
        'type': 'topic',
        'domains': ['engineering', 'technology']
    })
    assert resp.status_code == 201

    current_search.flush_and_refresh('communities')

    # testing search
    resp = client.get(list_url, query_string={'q': 'title:BLR'})
    assert resp.status_code == 200
    assert resp.json['hits']['hits'][0]['metadata']['id'] == 'biosyslit'
    assert resp.json['hits']['hits'][0]['metadata']['title'] == 'BLR'
    assert resp.json['hits']['hits'][0]['metadata']['type'] == 'topic'
    assert resp.json['hits']['hits'][0]['links']['self'] == 'http://localhost/communities/biosyslit'

    resp = client.get(list_url, query_string={'type': 'topic'})
    assert resp.status_code == 200
    assert resp.json['hits']['total'] == 2

    resp = client.get(list_url, query_string={'domain': ['engineering']})
    assert resp.status_code == 200
    assert resp.json['hits']['total'] == 2
    resp = client.get(list_url, query_string={'domain': ['technology']})
    assert resp.status_code == 200
    assert resp.json['hits']['total'] == 1
